/*
 * BasicEncoding.hpp
 *
 *  Created on: 21.07.2009
 *      Author: stephan
 */

#ifndef BASICENCODING_HPP_
#define BASICENCODING_HPP_

#include "Graph.hpp"
#include "BasicSolver.hpp"
#include <string>
using std::string;


class Var{
	uint id;
	string descr;
	Var(uint ID,string expl):id(ID),descr(expl){}

public:
	uint getID(){return id;}
	string getDescr(){return descr;}

	friend class Translator;
};


class Translator{
	vector<Var*> 		vars;

public:
	string get(uint i){
		assert(i < vars.size());
		Var*  trans = vars[i];
		assert(trans != NULL);
		return trans->getDescr();
	}

	string operator()(const Variable& v){
		string s = get(v.idx());
		return (s[0] =='!') ? s: "";}

	Var* newVar(string expl){
		assert(expl.size() != 0);
		Var* v = new Var(vars.size(),expl);
		vars.push_back(v);
		return v;
	}

	uint varsCount(){return vars.size();}


	~Translator(){for(uint i=vars.size();i--;){delete vars[i];}	}
};


template<class C>
class BasicNode{
	Vertex			graph_vertex;
	vector<C*>		possible_cells;
	vector<Var*>	possible_cells_vars;
public:
	vector<Var*>&   getCellVars()	{return possible_cells_vars;}
	uint 			getID()			{return graph_vertex.id;}
	Vertex 			getVertex()		{return graph_vertex;}
	BasicNode(const Vertex& v):graph_vertex(v){}
	void addPossibleCell(C* c,Var* v){
		possible_cells		.push_back(c);
		possible_cells_vars	.push_back(v);
	}
	Var* getConditionAtCell(C* c){
		for(uint i=0;i!=possible_cells.size();++i){
			if(possible_cells[i]==c) return possible_cells_vars[i];
		}
		return NULL;
	}
};

template<class N, class C>
class BasicCell{

	vector<N*> 		possible_nodes;
	vector<Var*>	possible_nodes_vars;

protected:
	uint row;
	uint col;

	void			addNodeMayBeEmptyConstraint(Var* empty){
					possible_nodes_vars.push_back(empty);}

	void			addPossibleNode(N* node, Var* var){
					assert(possible_nodes_vars.size() == possible_nodes.size());
					possible_nodes.push_back(node);
					possible_nodes_vars.push_back(var);}

public:
	BasicCell<N,C>(uint r, uint c):row(r),col(c){};
	const vector<Var*>&  getCardConstraintVars()	{return possible_nodes_vars;}
	N*   			getNode(uint at){
					assert(at < possible_nodes.size());assert(possible_nodes[at]!=NULL);return possible_nodes[at];}

	Var* 			getVar(uint at){
					assert(at < possible_nodes_vars.size());return possible_nodes_vars[at];}

	uint			getRow()		{return row;}
	uint			getColumn()		{return col;}
	uint			nodesCount()	{return possible_nodes.size();}

	Var* 			getConditionNodeAtThis(N* node){
					for(uint i=0;i != possible_nodes.size();++i){
						if(possible_nodes[i] == node){return possible_nodes_vars[i];}}
					return NULL;}

	Var* 			getConditionNoNodeAtThis(){
					assert(possible_nodes.size() <= possible_nodes_vars.size());
					return (possible_nodes.size() != possible_nodes_vars.size()) ?
							NULL : 	possible_nodes_vars.back();}
};




template<class N, class C>
class BasicEncoder{
protected:
	Graph& graph;
	BasicSolver& solver;
	Translator&  trans;

	uint width;
	uint height;

	uint vertex_count;
	uint edge_count;

	inline bool initGridSize(uint w, uint h);

	inline bool addABImpliesC(Var* a, Var* b, Var* c);
	inline bool addAImpliesB(Var* a, Var* b);
	inline bool addNotAOrNotB(Var* a, Var* b);
	inline bool addAtLeastOneFalse(Var* a, Var* b,Var* c,Var* d);
	inline bool addClause( const  vector<Var*>& vars, const vector<bool>& signs);
	inline bool addCardConstraint(const  vector<Var*>& vars);

	uint cellID(C* c){return c->getRow()*width + c->getColumn();}


	BasicEncoder(Graph& g,BasicSolver& s,Translator& translator)
	:	graph(g), solver(s),trans(translator),
		width(0), height(0), vertex_count(g.vertexCount()),	edge_count(g.edgeCount()){};

public:
	uint rowCount(){return height;}
	uint colCount(){return width;}
};

template<class N,class C>
bool BasicEncoder<N,C>::initGridSize(uint w, uint h){

	width 	= w == 0 ? graph.parsed_grid_width : w;
	height 	= h == 0 ? graph.parsed_grid_height : h;

	if(width == 0 && height == 0){
		width 	= (uint)sqrt(vertex_count);
		height 	= (uint)ceil((double)vertex_count / (double)width);
	}else if(width == 0){
		assert(height != 0);
		width 	= (uint)ceil((double)vertex_count / (double)height);
	}else if(height == 0){
		assert(width != 0);
		height 	= (uint)ceil((double)vertex_count / (double)width);
	}

	return width*height >= vertex_count;
}

template<class N,class C>
bool BasicEncoder<N,C>::addCardConstraint(const vector<Var*>& vars){
	vector<Literal> lits;

	for(uint i=0;i!= vars.size();++i){
		Variable v(vars[i]->getID());
		lits.push_back(v.toPosLit());
	}
	return solver.addCardinalityConstraint(lits);
}
/*
bool BasicEncoder::addClause(const  vector<Var*>& vars, const vector<bool>& signs){
	assert(vars.size()==signs.size());
	vector<Literal> lits;

	for(uint i=0;i!= vars.size();++i){
		Variable v(vars[i]->getID());
		lits.push_back(signs[i]?v.toPosLit():v.toNegLit());
	}
	return solver.addClause(lits);
}*/
template<class N,class C>
bool BasicEncoder<N,C>::addABImpliesC(Var* a, Var* b, Var* c){
	vector<Literal> lits;

	lits.push_back((Variable(a->getID())).toNegLit());
	lits.push_back((Variable(b->getID())).toNegLit());
	if(c != NULL){
		lits.push_back((Variable(c->getID())).toPosLit());
	}

	return solver.addClause(lits);
}
template<class N,class C>
bool BasicEncoder<N,C>::addAImpliesB(Var* a, Var* b){
	vector<Literal> lits;

	lits.push_back((Variable(a->getID())).toNegLit());
	if(b != NULL){
		lits.push_back((Variable(b->getID())).toPosLit());
	}

	return solver.addClause(lits);
}


template<class N,class C>
bool BasicEncoder<N,C>::addNotAOrNotB(Var* a, Var* b){
	vector<Literal> lits;

	lits.push_back((Variable(a->getID())).toNegLit());
	lits.push_back((Variable(b->getID())).toNegLit());

	return solver.addClause(lits);
}

template<class N,class C>
bool BasicEncoder<N,C>::addAtLeastOneFalse(Var* a, Var* b,Var* c,Var* d){
	vector<Literal> lits;

	lits.push_back((Variable(a->getID())).toNegLit());
	lits.push_back((Variable(b->getID())).toNegLit());
	lits.push_back((Variable(c->getID())).toNegLit());
	lits.push_back((Variable(d->getID())).toNegLit());

	return solver.addClause(lits);
}










#endif /* BASICENCODING_HPP_ */
