/*
 * CatActHeap.hpp
 *
 *  Created on: 08.06.2009
 *      Author: stephan
 */

#ifndef CATACTHEAP_HPP_
#define CATACTHEAP_HPP_

#include <stdint.h>
#include "MaxHeap.hpp"
#include "Random.hpp"
#include <utility>

using std::pair;

class Category{
	uint16_t	ID;
public:
	static const uint16_t INVALID_CAT = (1<<15) -1;

	explicit 	Category(uint id = INVALID_CAT)	:ID(id){}
	uint16_t 	idx()const			{return ID;}
	bool	 	operator==(const Category& c) const {return ID == c.ID;}
	bool		isValid() const{return ID != INVALID_CAT;}
};

class CatActHeap;

class CategoryPair{
	uint16_t cat1_pol;
	uint16_t cat2_pol;

	void setCategory(const Category& c, bool pol_true, bool first){
		if(first) 	cat1_pol = c.idx()<<1 | ((uint)pol_true);
		else 		cat2_pol = c.idx()<<1 | ((uint)pol_true);}

public:
	explicit CategoryPair(const Category& c1 = Category(), const Category& c2 = Category())
		:cat1_pol(c1.idx()<<1),cat2_pol(c2.idx()<<1){}


	bool categoriesed() 	const	{return (cat1_pol>>1) != Category::INVALID_CAT;}
	bool hasTwoCategories() const 	{return (cat2_pol>>1) != Category::INVALID_CAT;}

	Category getFirst() 	const 	{return Category((cat1_pol>>1));}
	Category getSecond() 	const 	{return Category(cat2_pol>>1);}

	bool polarity(const bool first_cat) const{
			return (first_cat ? cat1_pol : cat2_pol) & 1;}

	bool polarity(const Category& cat) const{
		assert(getFirst() == cat || getSecond() == cat);
		return polarity(getFirst() == cat);}

	bool twoCatsWithPolarity(bool pol) const{
		return 	hasTwoCategories()	&& 	(cat1_pol&1)==pol && (cat2_pol&1)==pol;}

	friend class CatActHeap;
};




class CategoryVar{
	uint32_t var31_c1;
public:
	CategoryVar(const Variable& v, bool first_cat):var31_c1((v.idx()<<1) | ((uint)first_cat) ){}
	uint32_t idx()		const	{return var31_c1;}
	Variable getVar() 	const 	{return Variable(var31_c1>>1);}
	bool	 operator==(const CategoryVar& v) const {return var31_c1 == v.var31_c1;}
};

struct CatVarActivityMap: public ActivityMap<Variable,var_act_t>{
	typedef ActivityMap<Variable,var_act_t> super;
	CatVarActivityMap(var_act_t MAX,var_act_t MIN, var_act_t dec): super(MAX,MIN,dec){}

	bool less(const CategoryVar& a, const CategoryVar& b){
		return super::less(a.getVar(),b.getVar());}

	void reserve(uint sz)			{super::reserve(sz);}
	void decayAll()					{super::decayAll();}
	bool bump(const Variable& v)	{return super::bump(v);}
};



class CatActHeap{

	typedef MaxHeap<Category, ActivityMap<Category,cat_act_t> , INTMap<Category> > 	cat_heap_t;
	typedef MaxHeap<CategoryVar, CatVarActivityMap , INTMap<CategoryVar>  > 		var_heap_t;

	//entry for each variable
	vector<CategoryPair> 				var_categories;
	CatVarActivityMap 					var_act_map;

	//two entries for each variable (since every var might have two categories)
	INTMap<CategoryVar>  				catvar_int_map;

	//for each category
	ActivityMap<Category,cat_act_t> 	cat_act_map;
	INTMap<Category>					cat_int_map;
	vector<var_heap_t>					var_heaps;

	cat_heap_t							cat_heap;
	//vector<Clause*>						cat_constraints;

	//heap for variables that don't belong to any category
	var_heap_t							var_default_hp;

	inline Variable	removeAtPos(const Category& cat, uint pos);


public:

	inline void 	reserveForVariables(uint32_t sz);

	inline void 	reserveForCategories(const Category& cat);

	inline bool		addElementIfNotExists(const Variable& variable);

	inline bool		setCategory(const Variable& variable, const Category& cat, bool pol_true );

	inline void		restoreVarAndCat(const Variable& v);

	inline bool 	bump(const Literal& lit, bool var_only_if_cat_match);


	const CategoryPair& getCats(const Variable& v){
					/*if(v.idx() >= var_categories.size()){
						print_out("Bad Variable Cat Request %d \n",v.toDIMACS());}*/
					assert(v.idx() < var_categories.size());
					return var_categories[v.idx()];}

	void			restoreCategory(const Category& cat){cat_heap.restoreElement(cat);}

	bool			empty()		{return cat_heap.empty() && var_default_hp.empty();}

	bool			empty(const Category& cat){
					assert(cat.idx() < var_heaps.size());
					return var_heaps[cat.idx()].empty();}

	bool			emptyDefaultCat(){
					return var_default_hp.empty();}

	Variable		removeRandomDefaultCat(Random& randy){
					return var_default_hp.removeRandom(randy).getVar();}

	Category 		removeRandomCategory(Random & randy){
					if(cat_heap.empty()) return Category();
					return cat_heap.removeRandom(randy);}

	Category 		removeMaxCategory(){
					if(cat_heap.empty()) return Category();
					return cat_heap.removeMax();}

	Variable		removeMaxDefaultCat(){
					return var_default_hp.removeMax().getVar();}

	Category 		getMaxCategory(){
					if(cat_heap.empty()) return Category();
					return cat_heap[0];}

	Category 		getRandomCategory(Random& randy){
					if(cat_heap.empty()) return Category();
					return cat_heap[randy.nextInt(cat_heap.size())];}

	void			decayAll(){
					cat_act_map.decayAll();
					var_act_map.decayAll();}

	Variable 		removeMax(const Category& cat) {
					return removeAtPos(cat, 0);}

	void 			removeCat(const Category& cat) {
					cat_heap.remove(cat);}

	void 			removeAll(const Category& cat) {
					assert(cat.idx() < var_heaps.size());
					var_heaps[cat.idx()].removeAll();}

	Variable 		removeRandom(const Category& cat, Random& randy){
					assert(cat.idx() < var_heaps.size());
					return removeAtPos(cat,randy.nextInt(var_heaps[cat.idx()].size()));}

	uint			size(Category& cat){
					assert(cat.idx() < var_heaps.size());
					return var_heaps[cat.idx()].size();	}

	Variable		get(Category& cat,uint pos){
					assert(pos < size(cat));
					return var_heaps[cat.idx()][pos].getVar();	}

	/*void			setCatConstraint(Category& cat,Clause* cls){
					assert(cls == NULL || cls->isCardConstraint());
					cat_constraints[cat.idx()] = cls;	}

	Clause* 		getCatConstraint(Category& cat){
					assert(cat.idx() < cat_constraints.size());
					return cat_constraints[cat.idx()];}*/

	Category 		determineMax(const Category& c1, const Category& c2){
					return cat_act_map.less(c1,c2) ? c2 : c1;}


	bool			varPolarityAtCat(const Variable& v, const Category& c){
					assert(v.idx() < var_categories.size());
					return var_categories[v.idx()].polarity(c);}

	/*bool			varHasCat(const Variable& v){
					assert(v.idx() < var_categories.size());
					return var_categories[v.idx()].categoriesed();}*/

	bool			contains(const Variable& v){
					assert(v.idx() < var_categories.size());
					return (catvar_int_map.get(CategoryVar(v,true)) != var_heap_t::HEAP_INVALID_INDEX);}

/*	bool			matchesPolarityAtCat(const Variable& v, bool pol, const Category& c){
					assert(v.idx() < var_categories.size());
					return var_categories[v.idx()].polarity(c) == pol;}
*/




	CatActHeap(var_act_t MAX,var_act_t MIN, var_act_t dec, cat_act_t maxc, cat_act_t minc, cat_act_t decc)
					:var_act_map(MAX,MIN,dec),
					catvar_int_map(),
					cat_act_map(maxc,minc,decc),
					cat_int_map(),
					cat_heap(&cat_act_map,&cat_int_map),
					var_default_hp(&var_act_map,&catvar_int_map){}

};

void CatActHeap::reserveForVariables(uint32_t sz){
	if(sz >var_categories.size()){
		var_categories	.resize(sz);
		var_act_map		.reserve(sz);
		catvar_int_map	.reserve(sz<<1,var_heap_t::HEAP_INVALID_INDEX);
	}
}

void CatActHeap::reserveForCategories(const Category& cat){
	assert(cat.isValid());
	uint32_t sz = cat.idx()+1;
	if(sz >var_heaps.size()){
		cat_int_map		.reserve(sz,cat_heap_t::HEAP_INVALID_INDEX);
		cat_act_map		.reserve(sz);
		var_heaps		.resize(sz,var_heap_t(&var_act_map,&catvar_int_map));
		//cat_constraints	.resize(sz,NULL);
	}
}



/*
bool CatActHeap::contains(const Variable& v){
	assert(	catvar_int_map.get(CategoryVar(v,true)) != var_heap_t::HEAP_INVALID_INDEX ||
			catvar_int_map.get(CategoryVar(v,false)) == var_heap_t::HEAP_INVALID_INDEX);
	return catvar_int_map.get(CategoryVar(v,true)) != var_heap_t::HEAP_INVALID_INDEX;
}*/

Variable CatActHeap::removeAtPos(const Category& cat, uint pos){
	assert(cat.idx() < var_heaps.size() && cat.isValid());
	var_heap_t& var_heap = var_heaps[cat.idx()];

	assert(!var_heap.empty() && pos < var_heap.size());
	return var_heap.removeAtPos(pos).getVar();
}

/*
Variable CatActHeap::removeRandom(Random& randy){
	if(cat_heap.empty()){
		assert(!var_default_hp.empty());
		return var_default_hp.removeRandom(randy).getVar();
	}
	Category cat = cat_heap[randy.nextInt(cat_heap.size())];
	return removeRandom(cat,randy);
}*/

bool	CatActHeap::setCategory(const Variable& v, const Category& cat, bool pol_true ){
	assert(v.idx() < var_categories.size());

	CategoryPair& cats = var_categories[v.idx()];
	if(cats.hasTwoCategories()) return false;

	reserveForCategories(cat);
	cat_heap.addElementIfNotExists(cat);

	bool first = true;
	CategoryVar catv(v,true);

	if(catvar_int_map.get(catv) != var_heap_t::HEAP_INVALID_INDEX) {
	//might be in default heap or in category 1 heap

		if(cats.categoriesed())	{first = false; }//in cat1heap so add to cat2
		else					{var_default_hp.remove(catv);}
	}

	cats						.setCategory(cat,pol_true,first);
	var_heaps[cat.idx()]		.addElementIfNotExists(CategoryVar(v,first));

	return true;
}

bool	CatActHeap::addElementIfNotExists(const Variable& v){
	assert(v.idx() < var_categories.size());
	CategoryVar cat1v(v,true);
	if(catvar_int_map.get(cat1v) != var_heap_t::HEAP_INVALID_INDEX) return false; //exists

	var_default_hp.addElementIfNotExists(cat1v);
	return true;
}


void	CatActHeap::restoreVarAndCat(const Variable& v){
	assert(v.idx() < var_categories.size());
	CategoryPair& cats = var_categories[v.idx()];
	CategoryVar cat1v(v,true);

	if(!cats.categoriesed()){
		var_default_hp.restoreElement(cat1v);
		return;
	}

	Category cat1 = cats.getFirst();
	assert(cat1.isValid() && cat1.idx() < var_heaps.size());
	cat_heap					.restoreElement(cat1);
	var_heaps[cat1.idx()]		.restoreElement(cat1v);


	Category cat2 = cats.getSecond();
	if(cat2.isValid()){
		assert(cat2.idx() < var_heaps.size());
		cat_heap				.restoreElement(cat2);
		var_heaps[cat2.idx()]	.restoreElement(CategoryVar(v,false));
	}
}

bool 	CatActHeap::bump(const Literal& lit,bool var_only_if_cat_match){
	Variable var(lit.toVar());
	assert(var.idx() < var_categories.size());
	CategoryPair& cats = var_categories[var.idx()];

	bool cheapb;
	CategoryVar cat1v(var,true);

	if(!cats.categoriesed()){
		cheapb = var_act_map	.bump(var);
		var_default_hp			.valueIncreased(cat1v);
		return cheapb;
	}

	Category cat1 = cats.getFirst();
	assert(cat1.isValid() && cat1.idx() < var_heaps.size());
	cheapb = cat_act_map		.bump(cat1);
	cat_heap					.valueIncreased(cat1);

	bool bump_var = !var_only_if_cat_match || (cats.polarity(true) == lit.isPos());

	Category cat2  = cats.getSecond();
	if(cat2.isValid()){
		assert(cat2.idx() < var_heaps.size());
		cheapb = cat_act_map	.bump(cat2) && cheapb;
		cat_heap				.valueIncreased(cat2);
		bump_var |=				(cats.polarity(false) == lit.isPos());
	}


	if(bump_var){
		cheapb = var_act_map		.bump(var) && cheapb;
		var_heaps[cat1.idx()]		.valueIncreased(cat1v);
		if(cat2.isValid()) {
			var_heaps[cat2.idx()]	.valueIncreased(CategoryVar(var,false));
		}
	}

	return cheapb;
}




#endif /* CATACTHEAP_HPP_ */
